package com.wechi.adweb.bridge.google.gtm.service;

import static com.wechi.adweb.bridge.google.common.ResourceDTO.parseId;

import com.google.api.client.googleapis.javanet.GoogleNetHttpTransport;
import com.google.api.client.json.gson.GsonFactory;
import com.google.api.client.util.Lists;
import com.google.api.services.tagmanager.TagManager;
import com.google.api.services.tagmanager.TagManagerScopes;
import com.google.api.services.tagmanager.model.*;
import com.google.auth.http.HttpCredentialsAdapter;
import com.google.auth.oauth2.GoogleCredentials;
import com.google.common.annotations.VisibleForTesting;
import com.wechi.adweb.bridge.exception.DataException;
import com.wechi.adweb.bridge.google.gtm.dto.*;
import com.wechi.adweb.bridge.util.JsonUtils;

import lombok.extern.slf4j.Slf4j;

import org.springframework.beans.factory.annotation.Value;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.security.GeneralSecurityException;
import java.time.LocalDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Collections;
import java.util.List;

import javax.annotation.PostConstruct;

/**
 * @author wfansh
 */
@Slf4j
@Service
public class GTMService {

    @Value("${google.service.account.key}")
    private String serviceAccountKey;

    @Value("${google.gtm.container.admins}")
    private List<String> containerAdmins;

    public TagManager tagManager;

    @PostConstruct
    private void init() throws GeneralSecurityException, IOException {
        // Service account authorization;
        GoogleCredentials credentials =
                GoogleCredentials.fromStream(
                                this.getClass()
                                        .getClassLoader()
                                        .getResourceAsStream(serviceAccountKey))
                        .createScoped(TagManagerScopes.all());

        this.tagManager =
                new TagManager.Builder(
                                GoogleNetHttpTransport.newTrustedTransport(),
                                GsonFactory.getDefaultInstance(),
                                new HttpCredentialsAdapter(credentials))
                        .build();
    }

    public List<GTMAccountDTO> listAccounts() throws DataException {
        try {
            List<GTMAccountDTO> gtmAccounts = Lists.newArrayList();
            for (Account account : tagManager.accounts().list().execute().getAccount()) {
                GTMAccountDTO gtmAccount =
                        GTMAccountDTO.builder()
                                .id(account.getAccountId())
                                .resourceName(account.getPath())
                                .displayName(account.getName())
                                .supportUserPermissions(
                                        account.getFeatures().getSupportUserPermissions())
                                .supportMultipleContainers(
                                        account.getFeatures().getSupportMultipleContainers())
                                .build();

                // Adds into the list;
                gtmAccounts.add(gtmAccount);
            }

            log.info("listAccounts : {}", JsonUtils.toJson(gtmAccounts));
            return gtmAccounts;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    public List<GTMContainerDTO> listContainers(String accountResourceName) throws DataException {
        try {
            List<GTMContainerDTO> gtmContainers = Lists.newArrayList();
            for (Container container :
                    tagManager
                            .accounts()
                            .containers()
                            .list(accountResourceName)
                            .execute()
                            .getContainer()) {
                GTMContainerDTO gtmContainer =
                        GTMContainerDTO.builder()
                                .id(container.getContainerId())
                                .resourceName(container.getPath())
                                .displayName(container.getName())
                                .accountId(container.getAccountId())
                                .domainNames(container.getDomainName())
                                .publicId(container.getPublicId())
                                .tagIds(container.getTagIds())
                                .notes(container.getNotes())
                                .usageContext(container.getUsageContext())
                                .fingerprint(container.getFingerprint())
                                .build();

                // Adds into the list;
                gtmContainers.add(gtmContainer);
            }

            log.info(
                    "listContainers for {} : {}",
                    accountResourceName,
                    JsonUtils.toJson(gtmContainers));
            return gtmContainers;
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    public GTMContainerDTO createContainer(String accountResourceName, String displayName)
            throws DataException {
        try {
            Container container = new Container();
            container.setName(displayName);
            container.setUsageContext(Collections.singletonList("web"));

            Container createdContainer =
                    tagManager
                            .accounts()
                            .containers()
                            .create(GTMAccountDTO.toResourceName(accountResourceName), container)
                            .execute();
            log.info("Created container {}", JsonUtils.toJson(createdContainer));

            // Converts and returns.
            return GTMContainerDTO.builder()
                    .id(createdContainer.getContainerId())
                    .resourceName(createdContainer.getPath())
                    .displayName(createdContainer.getName())
                    .accountId(parseId(accountResourceName))
                    .domainNames(createdContainer.getDomainName())
                    .publicId(createdContainer.getPublicId())
                    .tagIds(createdContainer.getTagIds())
                    .notes(createdContainer.getNotes())
                    .usageContext(createdContainer.getUsageContext())
                    .fingerprint(container.getFingerprint())
                    .build();
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    @VisibleForTesting
    void initContainerAccess(String accountResourceName, String containerId) throws DataException {
        try {
            // 1. Container access.
            // See
            // https://developers.google.com/tag-platform/tag-manager/api/v2/reference/accounts/user_permissions#containerAccess
            ContainerAccess containerAccess = new ContainerAccess();
            containerAccess.setContainerId(containerId);
            // publish > approve > edit > read
            containerAccess.setPermission("publish");

            // 2. Account access - required in user permission request.
            // See
            // https://developers.google.com/tag-platform/tag-manager/api/v2/reference/accounts/user_permissions#accountAccess
            AccountAccess accountAccess = new AccountAccess();
            accountAccess.setPermission("admin");

            // 3. Combines as user permission.
            UserPermission userPermission = new UserPermission();
            userPermission.setContainerAccess(Collections.singletonList(containerAccess));
            userPermission.setAccountAccess(accountAccess);

            for (String containerAdmin : containerAdmins) {
                // One email address per request.
                userPermission.setEmailAddress(containerAdmin);
                UserPermission createdUserPermission =
                        tagManager
                                .accounts()
                                .userPermissions()
                                .create(accountResourceName, userPermission)
                                .execute();
                log.info("Created user permission {}", JsonUtils.toJson(createdUserPermission));
            }
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    @VisibleForTesting
    GTMWorkspaceDTO getDefaultWorkspace(String containerResourceName) throws DataException {
        try {
            Workspace workspace =
                    tagManager
                            .accounts()
                            .containers()
                            .workspaces()
                            .list(containerResourceName)
                            .execute()
                            .getWorkspace()
                            .stream()
                            .filter(w -> w.getName().equalsIgnoreCase("Default Workspace"))
                            .findFirst()
                            .get();

            log.info(
                    "Default workspace for {} : {}",
                    containerResourceName,
                    JsonUtils.toJson(workspace));

            // Converts and returns.
            return GTMWorkspaceDTO.builder()
                    .id(workspace.getWorkspaceId())
                    .resourceName(workspace.getPath())
                    .displayName(workspace.getName())
                    .accountId(workspace.getAccountId())
                    .containerId(workspace.getContainerId())
                    .description(workspace.getDescription())
                    .fingerprint(workspace.getFingerprint())
                    .build();
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    @VisibleForTesting
    GTMTriggerDTO createTrigger(String workspaceResourceName) throws DataException {
        try {
            Trigger trigger = new Trigger();
            trigger.setName("Page View Trigger");
            // See
            // https://developers.google.com/tag-platform/tag-manager/api/v2/reference/accounts/containers/workspaces/triggers#type
            trigger.setType("pageview");

            Trigger createdTrigger =
                    tagManager
                            .accounts()
                            .containers()
                            .workspaces()
                            .triggers()
                            .create(workspaceResourceName, trigger)
                            .execute();
            log.info("Created trigger {}", JsonUtils.toJson(createdTrigger));

            // Converts and returns.
            return GTMTriggerDTO.builder()
                    .id(createdTrigger.getTriggerId())
                    .resourceName(createdTrigger.getPath())
                    .displayName(createdTrigger.getName())
                    .accountId(createdTrigger.getAccountId())
                    .containerId(createdTrigger.getContainerId())
                    .workspaceId(createdTrigger.getWorkspaceId())
                    .type(createdTrigger.getType())
                    .notes(createdTrigger.getNotes())
                    .fingerprint(createdTrigger.getFingerprint())
                    .build();
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    @VisibleForTesting
    void createGoogleTag(String workspaceResourceName, String triggerId, String googleTagId)
            throws DataException {
        try {
            Parameter parameter = new Parameter();
            // See
            // https://developers.google.com/tag-platform/tag-manager/api/v2/reference/accounts/containers/workspaces/tags#parameter.type
            parameter.setType("template");
            parameter.setKey("tagId");
            parameter.setValue(googleTagId);

            Tag tag = new Tag();
            tag.setName("Google Tag - PV");
            tag.setType("googtag");
            tag.setParameter(Collections.singletonList(parameter));
            tag.setFiringTriggerId(Collections.singletonList(triggerId));

            Tag createdTag =
                    tagManager
                            .accounts()
                            .containers()
                            .workspaces()
                            .tags()
                            .create(workspaceResourceName, tag)
                            .execute();
            log.info("Created tag {}", JsonUtils.toJson(createdTag));
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    @VisibleForTesting
    GTMVersionDTO createContainerVersion(String workspaceResourceName) throws DataException {
        try {
            CreateContainerVersionRequestVersionOptions versionOptions =
                    new CreateContainerVersionRequestVersionOptions();
            versionOptions.setName(
                    "Latest version #"
                            + LocalDateTime.now()
                                    .format(DateTimeFormatter.ofPattern("yyyyMMddHHmmss")));

            ContainerVersion createdContainerVersion =
                    tagManager
                            .accounts()
                            .containers()
                            .workspaces()
                            .createVersion(workspaceResourceName, versionOptions)
                            .execute()
                            .getContainerVersion();
            log.info("Created container version {}", JsonUtils.toJson(createdContainerVersion));

            // Converts and returns.
            return GTMVersionDTO.builder()
                    .id(createdContainerVersion.getContainerVersionId())
                    .resourceName(createdContainerVersion.getPath())
                    .displayName(createdContainerVersion.getName())
                    .accountId(createdContainerVersion.getAccountId())
                    .containerId(createdContainerVersion.getContainerId())
                    .deleted(createdContainerVersion.getDeleted())
                    .description(createdContainerVersion.getDescription())
                    .fingerprint(createdContainerVersion.getFingerprint())
                    .build();
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }

    @VisibleForTesting
    void publishContainerVersion(String versionResourceName) throws Exception {
        try {
            tagManager.accounts().containers().versions().publish(versionResourceName).execute();
            log.info("Published container version {}", JsonUtils.toJson(versionResourceName));
        } catch (IOException e) {
            log.error(e.getMessage());
            throw new DataException(e);
        }
    }
}
